#include <sys/cdefs.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
#include <inttypes.h>
#include <string.h>
#include <pthread.h>
#include <sys/queue.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <tcl.h>
#include <generic/tclInt.h>

#ifdef DMALLOC
#include "dmalloc.h"
#endif

#include <foo/core.h>
#include <foo/tcl.h>
#include <foo/snmpex.h>

#ifndef lint
static const char rcsid[] __unused = 
	"$Id: tclnetsnmp.c,v 1.41.2.12.4.13 2019/11/30 09:24:52 swp Exp $";
#endif /* !lint */

DEFN_ERROR(E_TCLNETSNMP_SESSIONMUTATION, "Attempt to change internal representation of an object netsnmp_session.")

static void NetsnmpSessionFree(Tcl_Obj *objPtr);
static void NetsnmpSessionDup(Tcl_Obj *objPtr, Tcl_Obj *copyPtr);
static void UpdateStringOfNetsnmpSession(Tcl_Obj *objPtr);
static int  SetNetsnmpSessionFromAny(Tcl_Interp *interp, Tcl_Obj *objPtr);

static const Tcl_ObjType tclNetsnmpSessionType = {
        "netsnmp_session",		/* name */
        NetsnmpSessionFree,		/* void freeIntRepProc(Tcl_Obj *objPtr)                     */
        NetsnmpSessionDup,		/* void dupIntRepProc(Tcl_Obj *objPtr, Tcl_Obj *copyPtr)    */
        UpdateStringOfNetsnmpSession,	/* void updateStringProc(Tcl_Obj *objPtr)                   */
        SetNetsnmpSessionFromAny	/* int  setFromAnyProc(Tcl_Interp *interp, Tcl_Obj *objPtr) */
};

#define TCLOBJ_NETSNMPSESSION_PTR(obj)	(*(netsnmp_session **)&(obj).internalRep.otherValuePtr)

/*     Если tcl хочет увидеть объект заданного типа (например list), то может
 * поменять внутреннее представление объекта (скажем на list) без оглядки на
 * кол-во ссылок (obj->refCount), т.е. даже если объект "shared".
 *     В нашем случае нельзя ни при каких условиях позволять менять внутреннее
 * представление. Если я правильно понимаю - это единственный способ дать tcl-у
 * по рукам.
 */
static
void
NetsnmpSessionFree(Tcl_Obj *obj)
{
#ifdef WITH_DEBUGLOG
        DLOG("%20s(): obj = <%p, %s, %d, <%p>>\n", __func__,
                obj, obj->typePtr ? obj->typePtr->name : "", obj->refCount,
                TCLOBJ_NETSNMPSESSION_PTR(*obj));
#endif
	if (obj->refCount <= 1) {
		if (TCLOBJ_NETSNMPSESSION_PTR(*obj)) {
			snmp_close(TCLOBJ_NETSNMPSESSION_PTR(*obj));
			TCLOBJ_NETSNMPSESSION_PTR(*obj) = NULL;
		}
		obj->typePtr = NULL;
	} else
		Tcl_Panic("[Tcl_Panic] %s(),%d: obj = <%p, %s, %d, <%p>>", 
			__func__, __LINE__, 
			obj, 
			tclobj_gettypename(obj),
			obj->refCount,
			TCLOBJ_NETSNMPSESSION_PTR(*obj));
}

static
void
NetsnmpSessionDup(Tcl_Obj *obj, Tcl_Obj *copy)
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	netsnmp_session session, *sess;
	int posix_errno, netsnmp_errno;
	char *err_message;

	ectlfr_begin(fr, L_panic_0);
	ectlno_begin(ex);
	sess = NULL;
	if (TCLOBJ_NETSNMPSESSION_PTR(*obj)) {
		snmp_sess_init(&session);
		session.peername = TCLOBJ_NETSNMPSESSION_PTR(*obj)->peername;
		sess = snmp_open(&session);
		if (!sess) {
			snmp_error(&session, &posix_errno, &netsnmp_errno, &err_message);
			ectlno_seterror(snmperr_to_error(netsnmp_errno));
			ectlno_printf("%s(),%d: {%s} snmp_open(): %s.\n", 
				__func__, __LINE__, error_name(ectlno_error), err_message);
			SNMP_FREE(err_message);
			ectlfr_goto(fr);
		}
		ectlfr_ontrap(fr, L_panic_1);
		sess->version = TCLOBJ_NETSNMPSESSION_PTR(*obj)->version;
		sess->timeout = TCLOBJ_NETSNMPSESSION_PTR(*obj)->timeout;
		free(sess->community);
		sess->community = (u_char *)STRDUP((char *)TCLOBJ_NETSNMPSESSION_PTR(*obj)->community);
		sess->community_len = TCLOBJ_NETSNMPSESSION_PTR(*obj)->community_len;
	}
	TCLOBJ_NETSNMPSESSION_PTR(*copy) = sess;
        copy->typePtr = &tclNetsnmpSessionType;
#ifdef WITH_DEBUGLOG
        DLOG("%20s(): obj = <%p, %s, %d, <%p>>, copy = <%p, %s, %d, <%p>>, bytes = %s, length = %d\n",
                __func__,
                 obj,  obj->typePtr ?  obj->typePtr->name : "",  obj->refCount, TCLOBJ_NETSNMPSESSION_PTR(*obj),
                copy, copy->typePtr ? copy->typePtr->name : "", copy->refCount, TCLOBJ_NETSNMPSESSION_PTR(*copy),
                obj->bytes, obj->length);
#endif
	ectlno_end(ex);
	ectlfr_end(fr);
	return;

L_panic_1:
	snmp_close(sess);
L_panic_0:
	ectlno_log();
	Tcl_Panic("[Tcl_Panic] %s,%d: BEGIN\n%s\nEND\n", __func__, __LINE__, ectlno_message);
}

static
void
UpdateStringOfNetsnmpSession(Tcl_Obj *obj)
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
        netsnmp_session *session;

	ectlfr_begin(fr, L_panic);
	ectlno_begin(ex);
	session = TCLOBJ_NETSNMPSESSION_PTR(*obj);
	if (!session) {
                if (obj->bytes != tclEmptyStringRep) {
                        if (obj->bytes != NULL)
                                ckfree((char *) obj->bytes);
                        obj->bytes = tclEmptyStringRep;
                        obj->length = 0;
                }
        } else {
		int n = strlen(session->peername) + 1 + 
			strlen(snmp_version_string(session->version)) + 1 +
				session->community_len;
		obj->bytes = CKREALLOC((obj->bytes != tclEmptyStringRep) ? obj->bytes : NULL, n + 1);
		snprintf(obj->bytes, n + 1, "%s %s %s", 
				session->peername, 
				snmp_version_string(session->version),
				session->community);
		obj->length = n;
	}
#ifdef WITH_DEBUGLOG
        DLOG("%20s(): obj = <%p, %s, %d, <%p>>, bytes = %s, length = %d\n",
                __func__, obj, obj->typePtr ? obj->typePtr->name : "", obj->refCount,
                TCLOBJ_NETSNMPSESSION_PTR(*obj), obj->bytes, obj->length);
#endif
	ectlno_end(ex);
	ectlfr_end(fr);
	return;

L_panic:
	ectlno_log();
	Tcl_Panic("[Tcl_Panic] %s,%d: BEGIN\n%s\nEND\n", __func__, __LINE__, ectlno_message);
}


static
int
SetNetsnmpSessionFromAny(Tcl_Interp *interp, Tcl_Obj *obj)
{
	struct ectlno ex[1];

	ectlno_begin(ex);
	ectlno_seterror(E_TCLNETSNMP_SESSIONMUTATION);
	ectlno_printf(
		"%s(),%d: forbidden operation.\n"
		"    obj = <%p, %s, %d>, bytes = %s, length = %d", 
		__func__, __LINE__, 
		obj, tclobj_gettypename(obj), obj->refCount, obj->bytes, obj->length);
	ectlno_tcl_setresult(interp);
        ectlno_clearmessage();
	ectlno_end(ex);
        return TCL_ERROR;
}


/* set sess [netsnmp::session ?-v <snmpVersion>? ?-c <snmpCommunity>? ?-? <snmpAgent>]
 * netsnmp::get $sess $vbl
 */
static
int
netsnmp_sessionObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	static const char usage[] = "netsnmp::session ?-(version|v) (1|2c)? ?-(community|c) community? "
				"?-(timeout|t) timeout? ?--|-? agent";
	int version = SNMP_VERSION_2c;
	const char *community = "public";
	long timeout = 0;
	netsnmp_session session, *sess = NULL;
	Tcl_Obj *obj;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	do {
		enum { xGET_OPT, xGET_OPT_ARG } x;
		enum { optVersion, optCommunity, optTimeout } optopt;
		const char *s;
		int i;

		x = xGET_OPT;
		for (i = 1; i < objc; i++) {
			s = Tcl_GetString(objv[i]);
			if (x == xGET_OPT) {
				if (*s != '-')
					break;
				s++;
				if (!*s || (*s == '-' && !s[1])) {
					i++;
					break;
				}
				if (!strcmp(s, "version") || !strcmp(s, "v")) {
					optopt = optVersion;
					x = xGET_OPT_ARG;
					continue;
				}
				if (!strcmp(s, "community") || !strcmp(s, "c")) {
					optopt = optCommunity;
					x = xGET_OPT_ARG;
					continue;
				}
				if (!strcmp(s, "timeout") || !strcmp(s, "t")) {
					optopt = optTimeout;
					x = xGET_OPT_ARG;
					continue;
				}
				ECTL_PTRAP(EINVAL, "Unknown option: -%s\n%s\n", s, usage);
			} else {
				if (optopt == optVersion) {
					if (!strcmp(s, "1"))
						version = SNMP_VERSION_1;
					else if (!strcmp(s, "2c"))
						version = SNMP_VERSION_2c;
					else
						ECTL_PTRAP(EINVAL, "Wrong/unsupported snmp version: %s.\n%s", s, usage);
				} else if (optopt == optCommunity) {
					community = s;
				} else if (optopt == optTimeout) {
					timeout = TCL_GETUINT32FROMOBJ(interp, objv[i]);
				}
				x = xGET_OPT;
			}
		}
		objc -= i;
		objv += i;
	} while (0);

	if (objc != 1)
		ECTL_PTRAP(EINVAL, "%s\n", usage);
	
	snmp_sess_init(&session);
	session.peername = Tcl_GetString(objv[0]);
	sess = snmp_open(&session);
	if (!sess) {
		int posix_errno, netsnmp_errno;
		char *err_message;

		snmp_error(&session, &posix_errno, &netsnmp_errno, &err_message);
		ectlno_seterror(snmperr_to_error(netsnmp_errno));
		ectlno_printf("%s(),%d: {%s} snmp_open(): %s.\n", 
			__func__, __LINE__, error_name(ectlno_error), err_message);
		SNMP_FREE(err_message);
		ectlfr_goto(fr);
	}
	ectlfr_ontrap(fr, L_1);
	sess->version = version;
	free(sess->community);
	sess->community = NULL;
	sess->community = (u_char *)STRDUP((char *)community);
	sess->community_len = strlen(community);
	if (timeout > 0)
		sess->timeout = timeout * 1000000L;

	obj = TCL_NEWOBJ();
	Tcl_InvalidateStringRep(obj);
	TCLOBJ_NETSNMPSESSION_PTR(*obj) = sess;
        obj->typePtr = &tclNetsnmpSessionType;
#ifdef WITH_DEBUGLOG
        DLOG("%s(),%d: obj = <%p, %s, %d, <%p>>, bytes = %s, length = %d\n", 
		__func__, __LINE__, obj, tclobj_gettypename(obj), obj->refCount, 
			TCLOBJ_NETSNMPSESSION_PTR(*obj), obj->bytes, obj->length);
#endif
	Tcl_SetObjResult(interp, obj);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_1:	ectlfr_ontrap(fr, L_0);
	snmp_close(sess);
L_0:	ectlno_tcl_setresult(interp);
        ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}
static
int
netsnmp_session_peernameObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc != 2)
		ECTL_PTRAP(EINVAL, "Wrong number of arguments.\n"
			"Usage: netsnmp::session.peername <netsnmp_session>.\n");
	if (objv[1]->typePtr != &tclNetsnmpSessionType)
		ECTL_PTRAP(EINVAL, "Wrong argument type for <netsnmp_session>: %s.\n",
				tclobj_gettypename(objv[1]));
	Tcl_SetResult(interp, TCLOBJ_NETSNMPSESSION_PTR(*objv[1])->peername, TCL_VOLATILE);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}
static
int
netsnmp_session_timeoutObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	long old_timeout;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc != 2 && objc != 3)
		ECTL_PTRAP(EINVAL, "Wrong number of arguments %d.\n"
			"Usage: netsnmp::session.timeout <netsnmp_session> ?<timeout>?.\n", objc);
	if (objv[1]->typePtr != &tclNetsnmpSessionType)
		ECTL_PTRAP(EINVAL, "Wrong argument type for <netsnmp_session>: %s.\n", 
			tclobj_gettypename(objv[1]));
	old_timeout = TCLOBJ_NETSNMPSESSION_PTR(*objv[1])->timeout / 1000000;
	if (objc != 2)
		TCLOBJ_NETSNMPSESSION_PTR(*objv[1])->timeout = TCL_GETUINT32FROMOBJ(interp, objv[2]) * 1000000;
	Tcl_SetObjResult(interp, TCL_NEWLONGOBJ(old_timeout));
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}


struct tcl_snmpvariable {
	Tcl_Obj *tv_name;
	Tcl_Obj *tv_index;
	Tcl_Obj *tv_type;
	Tcl_Obj *tv_value;
	Tcl_Obj *tv_enum;
};

static inline
void
tcl_snmpvariable_init(struct tcl_snmpvariable *tv)
{
	tv->tv_name = tv->tv_index = tv->tv_type = tv->tv_value = tv->tv_enum = NULL;
}

static
void
tcl_snmpvariable_fini(struct tcl_snmpvariable *tv)
{
	if (tv->tv_name) {
		Tcl_DecrRefCount(tv->tv_name);
		tv->tv_name = 0;
	}
	if (tv->tv_index) {
		Tcl_DecrRefCount(tv->tv_index);
		tv->tv_index = 0;
	}
	if (tv->tv_type) {
		Tcl_DecrRefCount(tv->tv_type);
		tv->tv_type = 0;
	}
	if (tv->tv_value) {
		Tcl_DecrRefCount(tv->tv_value);
		tv->tv_value = 0;
	}
	if (tv->tv_enum) {
		Tcl_DecrRefCount(tv->tv_enum);
		tv->tv_enum = 0;
	}
}

static
void
tcl_snmpvariable_set(Tcl_Interp *interp, struct tcl_snmpvariable *tv, Tcl_Obj *Index, netsnmp_variable_list *v)
{
	struct ectlfr fr[1];

	ectlfr_begin(fr, L_1);
	if (Index) {
		char name[100];
		snprint_objid(name, sizeof name, v->name, v->name_length - TCL_LISTOBJLENGTH(interp, Index));
		Tcl_IncrRefCount(Index);
		tv->tv_index = Index;
		tv->tv_name = tcl_get_string(name);
	} else {
		oid *oid_index;
		char name[100];
		struct tree *t = oid_find(get_tree_head(), v->name, v->name_length, &oid_index);
		if (!t) {
			snprint_objid(name, sizeof name, v->name, v->name_length);
			ECTL_TRAP(E_NOTFOUND, "oid_find(): %s: oid not found.\n", name);
		}
		snprint_objid(name, sizeof name, v->name, oid_index - v->name);
		tv->tv_index = TCL_NEWLISTOBJ(0, 0);
		Tcl_IncrRefCount(tv->tv_index);
		for (int i = v->name_length - (oid_index - v->name); i; i--)
			TCL_LISTOBJAPPENDELEMENT(interp, tv->tv_index, TCL_NEWLONGOBJ(*oid_index++));
		tv->tv_name = tcl_get_string(name);
	}

	if (v->type == ASN_INTEGER) {
		long val = *(long *)v->val.integer;
		tv->tv_value = TCL_NEWLONGOBJ(val);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("INTEGER");
		struct tree *t = get_tree(v->name, v->name_length, get_tree_head());
		if (t)
			for (struct enum_list *e = t->enums; e; e = e->next)
				if (e->value == val) {
					tv->tv_enum = tcl_get_string(e->label);
					break;
				}
	} else if (v->type == ASN_COUNTER) {
		tv->tv_value = TCL_NEWWIDEINTOBJ(*(u_long *)v->val.integer);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("COUNTER");
	} else if (v->type == ASN_GAUGE) {
		tv->tv_value = TCL_NEWWIDEINTOBJ(*(u_long *)v->val.integer);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("GAUGE");
	} else if (v->type == ASN_TIMETICKS) {
		tv->tv_value = TCL_NEWWIDEINTOBJ(*(u_long *)v->val.integer);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("TIMETICKS");
	} else if (v->type == ASN_UINTEGER) {
		u_long val = *(u_long *)v->val.integer;
		tv->tv_value = TCL_NEWWIDEINTOBJ(val);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("UINTEGER");
		struct tree *t = get_tree(v->name, v->name_length, get_tree_head());
		if (t)
			for (struct enum_list *e = t->enums; e; e = e->next)
				if (e->value == (int)val) {
					tv->tv_enum = tcl_get_string(e->label);
					break;
				}
	} else if (v->type == ASN_COUNTER64) {
		U64 *u64 = (U64 *)v->val.integer;
		union {
			uint64_t u64;
			uint32_t u32[2];
		} x;
		x.u32[0] = u64->low;
		x.u32[1] = u64->high;
		tv->tv_value = TCL_NEWWIDEINTOBJ(x.u64);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("COUNTER64");
	} else if (v->type == ASN_OCTET_STR) {
		int hexString = 0;
		const char *typecstr = "STRING";
		struct tree *t;
		const char *descriptor;

		t = get_tree(v->name, v->name_length, get_tree_head());
		descriptor = 0;
		if (t && t->tc_index != -1)
			descriptor = get_tc_descriptor(t->tc_index);
		do {
			if (descriptor) {
				/* DLOG("name: %s, descriptor: %s\n", Tcl_GetString(tv->tv_name), descriptor); */
				if (!strcmp(descriptor, "DisplayString")) {
					tv->tv_value = TCL_NEWSTRINGOBJ((char *)v->val.string, v->val_len);
					break;
				}
				if (!strcmp(descriptor, "MacAddress")) {
					char buf[18];
					snprintf(buf, sizeof buf, "%02x:%02x:%02x:%02x:%02x:%02x", 
						v->val.string[0], v->val.string[1], v->val.string[2],
						v->val.string[3], v->val.string[4], v->val.string[5]);
					tv->tv_value = TCL_NEWSTRINGOBJ(buf, sizeof buf - 1);
					break;
				}
				/* textual conventions */
			}
			for (size_t i = 0; i < v->val_len; i++)
				if (!isprint(v->val.string[i]) && !isspace(v->val.string[i])) {
					hexString = 1;
					break;
				}
			if (hexString) {
				char buf[v->val_len * 2 + 1];
				for (size_t i = 0; i < v->val_len; i++) {
					uint8_t b = v->val.string[i];
					buf[2*i] = "0123456789abcdef"[(b>>4) & 0x0f];
					buf[2*i + 1] = "0123456789abcdef"[b & 0x0f];
				}
				buf[v->val_len * 2] = 0;
				tv->tv_value = TCL_NEWSTRINGOBJ(buf, v->val_len * 2);
				typecstr = "Hex-STRING";
			} else
				tv->tv_value = TCL_NEWSTRINGOBJ((char *)v->val.string, v->val_len);
		} while (0);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string(typecstr);
	} else if (v->type == ASN_IPADDRESS) {
		char ip_str[16];
		uint8_t *ip = (uint8_t *)v->val.string;
		snprintf(ip_str, sizeof ip_str, "%u.%u.%u.%u", ip[0], ip[1], ip[2], ip[3]);
		tv->tv_value = TCL_NEWSTRINGOBJ(ip_str, -1);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("IPADDRESS");
	} else if (v->type == ASN_OPAQUE) {
		tv->tv_value = TCL_NEWSTRINGOBJ((char *)v->val.string, v->val_len);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("OPAQUE");
	} else if (v->type == ASN_NSAP) {
		tv->tv_value = TCL_NEWSTRINGOBJ((char *)v->val.string, v->val_len);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("NSAP");
	} else if (v->type == ASN_OBJECT_ID) {
		Tcl_DString ds[1];
		char buf[0x40];

		Tcl_DStringInit(ds);
		sprintf(buf, "%" PRIu32, (uint32_t)v->val.objid[0]);
		Tcl_DStringAppend(ds, buf, -1);
		for (size_t i = 1; i < v->val_len/sizeof(oid); i++) {
			sprintf(buf, ".%" PRIu32, (uint32_t)v->val.objid[i]);
			Tcl_DStringAppend(ds, buf, -1);
		}
		tv->tv_value = Tcl_NewStringObj(Tcl_DStringValue(ds), Tcl_DStringLength(ds));
		if (!tv->tv_value) {
			ectlno_setposixerror(ENOMEM);
			ectlno_printf("%s(),%d: {%s} Tcl_NewStringObj(\"%s\", %d) failed.\n", 
				__func__, __LINE__, error_name(ectlno_error), 
				Tcl_DStringValue(ds), Tcl_DStringLength(ds));
			Tcl_DStringFree(ds);
			ectlfr_goto(fr);
		}
		Tcl_DStringFree(ds);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("OID");
	} else if (v->type == ASN_BIT_STR) {
		Tcl_DString ds[1];

		Tcl_DStringInit(ds);
		for (size_t i = 0; i < v->val_len; i++) {
			uint8_t b = *(u_char *)(v->val.string + i);
			for (int j = 0; j < 8; j++)
				Tcl_DStringAppend(ds, (b & (128 >> j)) ? "1" : "0", 1);
		}
		tv->tv_value = Tcl_NewStringObj(Tcl_DStringValue(ds), Tcl_DStringLength(ds));
		if (!tv->tv_value) {
			ectlno_setposixerror(ENOMEM);
			ectlno_printf("%s(),%d: {%s} Tcl_NewStringObj(\"%s\", %d) failed.\n", 
				__func__, __LINE__, error_name(ectlno_error), 
				Tcl_DStringValue(ds), Tcl_DStringLength(ds));
			Tcl_DStringFree(ds);
			ectlfr_goto(fr);
		}
		Tcl_DStringFree(ds);
		Tcl_IncrRefCount(tv->tv_value);
		tv->tv_type = tcl_get_string("BIT_STR");
	} else if (v->type == ASN_NULL) {
		tv->tv_value = tcl_get_string("Null");
		tv->tv_type = tcl_get_string("NULL");
	} else
		ECTL_PTRAP(EINVAL, "Unknown snmp type: %lu.\n", (u_long)v->type);
	ectlfr_end(fr);
	return;

L_1:	ectlfr_ontrap(fr, L_0);
	tcl_snmpvariable_fini(tv);
L_0:	ectlfr_end(fr);
	ectlfr_trap();
}

static
void
tcl_snmpvariable_add_to_dict(Tcl_Interp *interp, Tcl_Obj *d, struct tcl_snmpvariable *tv)
{
	struct {
		const char *tvs_cname;
		int tvs_optional;
		Tcl_Obj **tvs_ppval;
	} tvs[] = {
		{ .tvs_cname = "name",  .tvs_optional = 0, .tvs_ppval = &tv->tv_name  },
		{ .tvs_cname = "index", .tvs_optional = 0, .tvs_ppval = &tv->tv_index },
		{ .tvs_cname = "type",  .tvs_optional = 0, .tvs_ppval = &tv->tv_type  },
		{ .tvs_cname = "value", .tvs_optional = 0, .tvs_ppval = &tv->tv_value },
		{ .tvs_cname = "enum",  .tvs_optional = 1, .tvs_ppval = &tv->tv_enum  }
	};
	Tcl_Obj *keyv[3];

	keyv[0] = tv->tv_index;
	keyv[1] = tv->tv_name;
	for (size_t i = 0; i < sizeof tvs/sizeof tvs[0]; i++) {
		if (tvs[i].tvs_optional && !*tvs[i].tvs_ppval)
			continue;
		keyv[2] = tcl_get_string(tvs[i].tvs_cname);
		Tcl_DecrRefCount(keyv[2]);
		TCL_DICTOBJPUTKEYLIST(interp, d, 3, keyv, *tvs[i].tvs_ppval);
	}
}

static
Tcl_Obj *
tcl_vbtree_to_dict(Tcl_Interp *interp, struct rbtree *vbtree)
{
	struct ectlfr fr[1];
	struct rbglue *gi, *gv;
	Tcl_Obj *volatile dict, *volatile index;
	struct tcl_snmpvariable tv[1];

	dict = TCL_NEWDICTOBJ();
	ectlfr_begin(fr, L_1);
        RBTREE_FOREACH(gi, vbtree) {
                struct vbtree_index *idp = (struct vbtree_index *)rbglue_dptr(gi);
		index = TCL_NEWLISTOBJ(0, 0);
		ectlfr_ontrap(fr, L_2);
		Tcl_IncrRefCount(index);
		for (size_t i = 0; i < idp->index_length; i++)
			TCL_LISTOBJAPPENDELEMENT(interp, index, TCL_NEWLONGOBJ(idp->index[i]));
                RBTREE_FOREACH(gv, idp->vbtree) {
                        netsnmp_variable_list *v = (netsnmp_variable_list *)rbglue_dptr(gv);
			tcl_snmpvariable_init(tv);
			ectlfr_ontrap(fr, L_3);
			tcl_snmpvariable_set(interp, tv, index, v);
			tcl_snmpvariable_add_to_dict(interp, dict, tv);
			ectlfr_ontrap(fr, L_2);
			tcl_snmpvariable_fini(tv);
                }
		ectlfr_ontrap(fr, L_1);
		Tcl_DecrRefCount(index);
        }
	ectlfr_end(fr);
	return dict;

L_3:	ectlfr_ontrap(fr, L_2);
	tcl_snmpvariable_fini(tv);
L_2:	ectlfr_ontrap(fr, L_1);
	Tcl_DecrRefCount(index);
L_1:	ectlfr_ontrap(fr, L_0);
	Tcl_DecrRefCount(dict);
L_0:	ectlfr_end(fr);
	ectlfr_trap();
}


static
int
netsnmp_getObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	struct rbtree *volatile result;
	int nvarnames;
	char **varnames;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc != 3)
		ECTL_PTRAP(EINVAL, "Wrong anount of arguments.\n"
			"Usage: netsnmp::get <netsnmp_session> <oid_list>.\n");
	TCLOBJ_CHECKTYPE(objv[1], &tclNetsnmpSessionType);
	nvarnames = TCL_LISTOBJLENGTH(interp, objv[2]);
	varnames = alloca(nvarnames * sizeof(char *));
	for (int i = 0; i < nvarnames; i++)
		varnames[i] = Tcl_GetString(TCL_LISTOBJINDEX(interp, objv[2], i));
	result = snmpget(TCLOBJ_NETSNMPSESSION_PTR(*objv[1]), varnames, nvarnames);
	/* vbtree_print(result); */
	ectlfr_ontrap(fr, L_1);

	Tcl_SetObjResult(interp, tcl_vbtree_to_dict(interp, result));
	ectlfr_ontrap(fr, L_0);
	vbtree_destroy(result);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_1:	ectlfr_ontrap(fr, L_1);
	vbtree_destroy(result);
L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}

static
int
netsnmp_walkObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	struct rbtree *result = NULL;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc != 3)
		ECTL_PTRAP(EINVAL, "Wrong amount of argumets.\n"
				"Usage: netsnmp::walk <netsnmp_session> <oid>.\n");
	TCLOBJ_CHECKTYPE(objv[1], &tclNetsnmpSessionType);
	result = snmpwalk(TCLOBJ_NETSNMPSESSION_PTR(*objv[1]), Tcl_GetString(objv[2]));
	/* vbtree_print(result); */
	ectlfr_ontrap(fr, L_1);
	Tcl_SetObjResult(interp, tcl_vbtree_to_dict(interp, result));
	ectlfr_ontrap(fr, L_0);
	vbtree_destroy(result);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_1:	ectlfr_ontrap(fr, L_1);
	vbtree_destroy(result);
L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}

static
int
netsnmp_bulkgetObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	static const char usage[] = "netsnmp::bulkget <netsnmp_session> ?-non-repeaters|-n n? "
					"?-max-repetitions|-r n? ?--|-? <oid_list>";
	netsnmp_session *session;
	int non_rep, max_rep, nvarnames;
	char **varnames;
	struct rbtree *volatile result;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc < 3)
		ECTL_PTRAP(EINVAL, "Wrong amount of argumets.\nUsage: %s\n", usage);
	TCLOBJ_CHECKTYPE(objv[1], &tclNetsnmpSessionType);
	session = TCLOBJ_NETSNMPSESSION_PTR(*objv[1]);
	non_rep = 0;
	max_rep = 16;
	do {
		enum { xGET_OPT, xGET_OPT_ARG, xOPT_END, xPARAMETERS } x;
		enum { optNonRepeaters, optMaxRepetitions } optopt;
		int i;

		x = xGET_OPT;
		for (i = 2; i < objc; i++) {
			const char *s = Tcl_GetString(objv[i]);
			if (x == xGET_OPT) {
				if (!strcmp(s, "-non-repeaters") || !strcmp(s, "-n")) {
					optopt = optNonRepeaters;
					x = xGET_OPT_ARG;
				} else if (!strcmp(s, "-max-repetitions") || !strcmp(s, "-r")) {
					optopt = optMaxRepetitions;
					x = xGET_OPT_ARG;
				} else if (*s == '-') {
					if (!s[1] || (s[1] == '-' && !s[2]))
						x = xOPT_END;
					else
						ECTL_PTRAP(EINVAL, "Unknown option: \"%s\".\n", s);
				} else {
					x = xPARAMETERS;
					break;
				}
			} else if (x == xGET_OPT_ARG) {
				if (optopt == optNonRepeaters) {
					if (ectlfr_catch(fr, non_rep = TCL_GETINTFROMOBJ(interp, objv[i]))) {
						ectlno_printf("%s(),%d: {%s} expected number for -non-repeaters.\n", 
							__func__, __LINE__, error_name(ectlno_error)); 
						ectlfr_goto(fr);
					}
				} else if (optopt == optMaxRepetitions) {
					if (ectlfr_catch(fr, max_rep = TCL_GETINTFROMOBJ(interp, objv[i]))) {
						ectlno_printf("%s(),%d: {%s} expected number for -max-repetitions.\n", 
							__func__, __LINE__, error_name(ectlno_error));
						ectlfr_goto(fr);
					}
				}
				x = xGET_OPT;
			} else if (x == xOPT_END) {
				x = xPARAMETERS;
				break;
			}
		}
		objc -= i;
		objv += i;
	} while (0);
	if (objc != 1)
		ECTL_PTRAP(EINVAL, "Wrong amount of arguments.\nUsage: %s.\n", usage);
	nvarnames = TCL_LISTOBJLENGTH(interp, objv[0]);
	if (nvarnames < non_rep)
		ECTL_TRAP(E_GENERIC, "Insufficient quantity of elements in the list of oids.\n"
				"Usage: %s.\n", usage);
	varnames = alloca(nvarnames * sizeof(char *));
	for (int i = 0; i < nvarnames; i++)
		varnames[i] = Tcl_GetString(TCL_LISTOBJINDEX(interp, objv[0], i));
	result = snmpbulkget(session, non_rep, max_rep, varnames, nvarnames);
	/* vbtree_print(result); */
	ectlfr_ontrap(fr, L_1);
	Tcl_SetObjResult(interp, tcl_vbtree_to_dict(interp, result));
	ectlfr_ontrap(fr, L_0);
	vbtree_destroy(result);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_1:	ectlfr_ontrap(fr, L_1);
	vbtree_destroy(result);
L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}

static
int
netsnmp_bulkwalkObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	static const char usage[] = 
		"netsnmp::bulkwalk <netsnmp_session> ?-non-repeaters|-n n? ?-max-repetitions|-r n? ?--|-? <oid_list>";
	int nvarnames;
	char **varnames;
	struct rbtree *volatile result;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc < 3)
		ECTL_PTRAP(EINVAL, "Wrong amount of argumets.\nUsage: %s.\n", usage);
	TCLOBJ_CHECKTYPE(objv[1], &tclNetsnmpSessionType);
	netsnmp_session *session = TCLOBJ_NETSNMPSESSION_PTR(*objv[1]);
	int non_rep = 0;
	int max_rep = 16;
	do {
		enum { xGET_OPT, xGET_OPT_ARG, xOPT_END, xPARAMETERS } x;
		enum { optNonRepeaters, optMaxRepetitions } optopt;
		int i;

		x = xGET_OPT;
		for (i = 2; i < objc; i++) {
			const char *s = Tcl_GetString(objv[i]);
			if (x == xGET_OPT) {
				if (!strcmp(s, "-non-repeaters") || !strcmp(s, "-n")) {
					optopt = optNonRepeaters;
					x = xGET_OPT_ARG;
				} else if (!strcmp(s, "-max-repetitions") || !strcmp(s, "-r")) {
					optopt = optMaxRepetitions;
					x = xGET_OPT_ARG;
				} else if (*s == '-') {
					if (!s[1] || (s[1] == '-' && !s[2]))
						x = xOPT_END;
					else
						ECTL_PTRAP(EINVAL, "Unknown option: \"%s\".\n", s);
				} else {
					x = xPARAMETERS;
					break;
				}
			} else if (x == xGET_OPT_ARG) {
				if (optopt == optNonRepeaters) {
					if (ectlfr_catch(fr, non_rep = TCL_GETINTFROMOBJ(interp, objv[i]))) {
						ectlno_printf("%s(),%d: {%s} expected number for -non-repeaters.\n", 
							__func__, __LINE__, error_name(ectlno_error)); 
						ectlfr_goto(fr);
					}
				} else if (optopt == optMaxRepetitions) {
					if (ectlfr_catch(fr, max_rep = TCL_GETINTFROMOBJ(interp, objv[i]))) {
						ectlno_printf("%s(),%d: {%s} expected number for -max-repetitions.\n", 
							__func__, __LINE__, error_name(ectlno_error));
						ectlfr_goto(fr);
					}
				}
				x = xGET_OPT;
			} else if (x == xOPT_END) {
				x = xPARAMETERS;
				break;
			}
		}
		objc -= i;
		objv += i;
	} while (0);
	if (objc != 1)
		ECTL_PTRAP(EINVAL, "Wrong amount of arguments.\nUsage: %s.\n", usage);
	nvarnames = TCL_LISTOBJLENGTH(interp, objv[0]);
	if (nvarnames < non_rep)
		ECTL_PTRAP(EINVAL, "Insufficient quantity of elements in the list of oids.\n"
				"Usage: %s.\n", usage);
	varnames = alloca(nvarnames * sizeof(char *));
	for (int i = 0; i < nvarnames; i++)
		varnames[i] = Tcl_GetString(TCL_LISTOBJINDEX(interp, objv[0], i));
	result = snmpbulkwalk(session, non_rep, max_rep, varnames, nvarnames);
	/* vbtree_print(result); */
	ectlfr_ontrap(fr, L_1);

	Tcl_SetObjResult(interp, tcl_vbtree_to_dict(interp, result));
	ectlfr_ontrap(fr, L_0);
	vbtree_destroy(result);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_1:	ectlfr_ontrap(fr, L_1);
	vbtree_destroy(result);
L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}

static
int
netsnmp_bulkwalk_exObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	struct ectlfr fr[1];
	struct ectlno ex[1];
	static const char usage[] = "netsnmp::bulkwalk_ex <netsnmp_session> ?-non-repeaters|-n n? "
				"?-max-repetitions|-r n? ?--|-? <oid_list>";
	int nvarnames;
	char **varnames;
	struct rbtree *volatile result;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc < 3)
		ECTL_PTRAP(EINVAL, "Wrong amount of argumets.\nUsage: %s.\n", usage);
	TCLOBJ_CHECKTYPE(objv[1], &tclNetsnmpSessionType);
	netsnmp_session *session = TCLOBJ_NETSNMPSESSION_PTR(*objv[1]);
	int non_rep = 0;
	int max_rep = 16;
	do {
		enum { xGET_OPT, xGET_OPT_ARG, xOPT_END, xPARAMETERS } x;
		enum { optNonRepeaters, optMaxRepetitions } optopt;
		int i;

		x = xGET_OPT;
		for (i = 2; i < objc; i++) {
			const char *s = Tcl_GetString(objv[i]);
			if (x == xGET_OPT) {
				if (!strcmp(s, "-non-repeaters") || !strcmp(s, "-n")) {
					optopt = optNonRepeaters;
					x = xGET_OPT_ARG;
				} else if (!strcmp(s, "-max-repetitions") || !strcmp(s, "-r")) {
					optopt = optMaxRepetitions;
					x = xGET_OPT_ARG;
				} else if (*s == '-') {
					if (!s[1] || (s[1] == '-' && !s[2]))
						x = xOPT_END;
					else
						ECTL_PTRAP(EINVAL, "Unknown option: \"%s\".\n", s);
				} else {
					x = xPARAMETERS;
					break;
				}
			} else if (x == xGET_OPT_ARG) {
				if (optopt == optNonRepeaters) {
					if (ectlfr_catch(fr, non_rep = TCL_GETINTFROMOBJ(interp, objv[i]))) {
						ectlno_printf("%s(),%d: {%s} expected number for -non-repeaters.\n", 
							__func__, __LINE__, error_name(ectlno_error)); 
						ectlfr_goto(fr);
					}
				} else if (optopt == optMaxRepetitions) {
					if (ectlfr_catch(fr, max_rep = TCL_GETINTFROMOBJ(interp, objv[i]))) {
						ectlno_printf("%s(),%d: {%s} expected number for -max-repetitions.\n", 
							__func__, __LINE__, error_name(ectlno_error));
						ectlfr_goto(fr);
					}
				}
				x = xGET_OPT;
			} else if (x == xOPT_END) {
				x = xPARAMETERS;
				break;
			}
		}
		objc -= i;
		objv += i;
	} while (0);
	if (objc != 1)
		ECTL_PTRAP(EINVAL, "Wrong amount of arguments.\nUsage: %s.\n", usage);
	nvarnames = TCL_LISTOBJLENGTH(interp, objv[0]);
	if (nvarnames < non_rep)
		ECTL_PTRAP(EINVAL, "Insufficient quantity of elements in the list of oids.\n"
				"Usage: %s.\n", usage);
	varnames = alloca(nvarnames * sizeof(char *));
	for (int i = 0; i < nvarnames; i++)
		varnames[i] = Tcl_GetString(TCL_LISTOBJINDEX(interp, objv[0], i));
	result = snmpbulkwalk_ex(session, non_rep, max_rep, varnames, nvarnames);
	/* vbtree_print(result); */
	ectlfr_ontrap(fr, L_1);
	Tcl_SetObjResult(interp, tcl_vbtree_to_dict(interp, result));
	ectlfr_ontrap(fr, L_0);
	vbtree_destroy(result);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_1:	ectlfr_ontrap(fr, L_1);
	vbtree_destroy(result);
L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}


static
int
netsnmp_setObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	static const char usage[] = "netsnmp::set <netsnmp_session> <oid> = <val> [oid = val [...]]";

	struct ectlfr fr[1];
	struct ectlno ex[1];
	netsnmp_session *session;
	netsnmp_pdu *volatile pdu, *volatile response;
	int status;
	char *name, type, *value;
	oid oid[MAX_OID_LEN];
	size_t oidlen = MAX_OID_LEN;

	ectlfr_begin(fr, L_0);
	ectlno_begin(ex);
	if (objc < 5 || ((objc-2) % 3) != 0)
		ECTL_PTRAP(EINVAL, "Wrong amount of argumets.\nUsage: %s.\n", usage);
	TCLOBJ_CHECKTYPE(objv[1], &tclNetsnmpSessionType);
	session = TCLOBJ_NETSNMPSESSION_PTR(*objv[1]);
	objc -= 2;
	objv += 2;
	pdu = snmp_pdu_create(SNMP_MSG_SET);
	if (!pdu)
		ECTL_PTRAP(ENOMEM, "snmp_pdu_create(SNMP_MSG_SET): %s.\n", strerror(ENOMEM));
	ectlfr_ontrap(fr, L_1);
	for (int i = 0; i < objc; i += 3) {
		name = Tcl_GetString(objv[i]);
		oidlen = MAX_OID_LEN;
		if (!snmp_parse_oid(name, oid, &oidlen))
			ECTL_PTRAP(EINVAL, "snmp_parse_oid(%s): %s.\n", name, snmp_api_errstring(snmp_errno));
		type = *Tcl_GetString(objv[i + 1]);
		value = Tcl_GetString(objv[i + 2]);
		if (snmp_add_var(pdu, oid, oidlen, type, value))
			ECTL_TRAP(E_GENERIC, "snmp_add_var(): Unable add variable \"%s\" \"%c\" \"%s\" to PDU.\n",
				name, type, value);
	}
        response = NULL;
        status = snmp_synch_response(session, pdu, (netsnmp_pdu **)&response);
        ectlfr_ontrap(fr, L_2);

        if (status == STAT_SUCCESS) {
                int errstat = response->errstat;
                if (errstat != SNMP_ERR_NOERROR) {
                        ectlno_setposixerror(EIO);
                        ectlno_printf("%s(),%d: {%s} Error in packet: \"%s\".\n",
                                __func__, __LINE__, error_name(ectlno_error), snmp_errstring(errstat));
                        if (response->errindex) {
                                ectlno_printf("\tFailed object: \"");
                                int i = 0;
                                for (netsnmp_variable_list *v = response->variables; v; v = v->next_variable)
                                        if (++i == response->errindex) {
                                                ectlno_print_oid(v->name, v->name_length);
                                                break;
                                        }
                                ectlno_printf("\".\n");
                        }
                        ectlfr_goto(fr);
                }
#if 0
        } else if (status == STAT_TIMEOUT) {
                ectlno_setposixerror(ETIMEDOUT);
                ectlno_printf("%s(),%d: {%s} snmp_synch_response(): Timeout: No Response from %s.\n",
                        __func__, __LINE__, error_name(ectlno_error), session->peername);
                ectlfr_goto(fr);
#endif
        } else {
		int posix_errno, netsnmp_errno;
                char *errstr;

                snmp_error(session, &posix_errno, &netsnmp_errno, &errstr);
		ectlno_seterror(snmperr_to_error(netsnmp_errno));
                ectlno_printf("%s(),%d: {%s} snmp_synch_response(): %s: %s.\n",
                        __func__, __LINE__, error_name(ectlno_error), session->peername, errstr);
                SNMP_FREE(errstr);
                ectlfr_goto(fr);
        }

        snmp_free_pdu(response);
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_OK;

L_2:    ectlfr_ontrap(fr, L_0);
        if (response)
                snmp_free_pdu(response);
        goto L_0;
L_1:	ectlfr_ontrap(fr, L_1);
	snmp_free_pdu(pdu);
L_0:	ectlno_tcl_setresult(interp);
	ectlno_clearmessage();
	ectlno_end(ex);
	ectlfr_end(fr);
	return TCL_ERROR;
}

static
int
netsnmp_add_mibdirObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "directory");
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}
	add_mibdir(Tcl_GetString(objv[1]));
	return TCL_OK;
}

#if !defined(NO_NETSNMP_READ_MODULE)
static
int
netsnmp_read_moduleObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "module-name");
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}
	netsnmp_read_module(Tcl_GetString(objv[1]));
	return TCL_OK;
}
#endif

static
int
netsnmp_read_mibObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "mib-file");
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}
	read_mib(Tcl_GetString(objv[1]));
	return TCL_OK;
}

static
int
netsnmp_read_all_mibsObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	if (objc != 1) {
		Tcl_WrongNumArgs(interp, 1, objv, "<no args>");
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}
	read_all_mibs();
	return TCL_OK;
}

static
int
netsnmp_oidObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "name");
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}

	oid oidbuf[MAX_OID_LEN];
	size_t oidlen = MAX_OID_LEN;
	char const *name = Tcl_GetString(objv[1]);

	if (!read_objid(name, oidbuf, &oidlen)) {
		Tcl_AppendResult(interp, name, ": ", snmp_api_errstring(snmp_errno), NULL);
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}

	Tcl_Obj *r = Tcl_ObjPrintf("%" NETSNMP_PRIo "u", oidbuf[0]);
	if (!r)
		return TCL_ERROR;
	Tcl_IncrRefCount(r);

	for (size_t i = 1; i < oidlen; i++)
		Tcl_AppendPrintfToObj(r, ".%" NETSNMP_PRIo "u", oidbuf[i]);

	Tcl_SetObjResult(interp, r);
	Tcl_DecrRefCount(r);
	return TCL_OK;
}

static
int
netsnmp_nameObjCmd(ClientData clientData __unused, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) 
{
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "Oid");
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}

	oid oidbuf[MAX_OID_LEN];
	size_t oidlen = MAX_OID_LEN;
	if (!snmp_parse_oid(Tcl_GetString(objv[1]), oidbuf, &oidlen)) {
		Tcl_AppendResult(interp, Tcl_GetString(objv[1]), ": ", snmp_api_errstring(snmp_errno), ".\n", NULL);
		Tcl_SetErrorCode(interp, "E_POSIX_EINVAL", NULL);
		return TCL_ERROR;
	}
	char name[512];
	snprint_objid(name, sizeof name, oidbuf, oidlen);
	Tcl_SetResult(interp, name, TCL_VOLATILE);
	return TCL_OK;
}

static struct {
        Tcl_Command token;
        const char *cmdName;
        Tcl_ObjCmdProc *proc;
        ClientData clientData;
        Tcl_CmdDeleteProc *deleteProc;
} cmd[] = {
	{
		.token = NULL, 
		.cmdName = "netsnmp::session", 
		.proc = netsnmp_sessionObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::session.peername", 
		.proc = netsnmp_session_peernameObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::session.timeout", 
		.proc = netsnmp_session_timeoutObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::get", 
		.proc = netsnmp_getObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::walk", 
		.proc = netsnmp_walkObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::bulkget", 
		.proc = netsnmp_bulkgetObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::bulkwalk", 
		.proc = netsnmp_bulkwalkObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::bulkwalk_ex", 
		.proc = netsnmp_bulkwalk_exObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::set", 
		.proc = netsnmp_setObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::add_mibdir", 
		.proc = netsnmp_add_mibdirObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
#if !defined(NO_NETSNMP_READ_MODULE)
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::read_module", 
		.proc = netsnmp_read_moduleObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
#endif
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::read_mib", 
		.proc = netsnmp_read_mibObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::read_all_mibs", 
		.proc = netsnmp_read_all_mibsObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::oid", 
		.proc = netsnmp_oidObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}, { 
		.token = NULL, 
		.cmdName = "netsnmp::name", 
		.proc = netsnmp_nameObjCmd, 
		.clientData = NULL, 
		.deleteProc = NULL 
	}
};

static
void
exit_handler(ClientData clientData __unused)
{
        Tcl_Interp *interp;

        interp = (Tcl_Interp *)clientData;
        for (size_t i = 0; i < sizeof cmd/sizeof cmd[0]; i++)
                Tcl_DeleteCommandFromToken(interp, cmd[i].token);
        Tcl_Release(interp);
	snmp_close_sessions();
	snmp_shutdown("snmpapp");
}


int PACKAGE_INIT(Tcl_Interp *interp)
{
        struct ectlfr fr[1];
        struct ectlno ex[1];

        ectlfr_begin(fr, L_0);
        ectlno_begin(ex);
        if (!Tcl_InitStubs(interp, __XSTRING(TCLVERSION), 0))
                ECTL_TRAP(E_GENERIC, "Tcl_InitStubs(): %s.\n", Tcl_GetStringResult(interp));
	init_snmp("snmpapp");
	ectlfr_ontrap(fr, L_1);
        Tcl_Preserve(interp);
	ectlfr_ontrap(fr, L_3);
        for (size_t i = 0; i < sizeof cmd/sizeof cmd[0]; i++) {
                cmd[i].token = Tcl_CreateObjCommand(interp, cmd[i].cmdName, cmd[i].proc,
                                        cmd[i].clientData, cmd[i].deleteProc);
                if (!cmd[i].token)
			ECTL_TRAP(E_GENERIC, "Tcl_CreateObjCommand(%s): %s.\n",
                                        cmd[i].cmdName, Tcl_GetStringResult(interp));
        }
        Tcl_CreateExitHandler(exit_handler, (ClientData)interp);
	Tcl_PkgProvide(interp, __XSTRING(PACKAGE), __XSTRING(PACKAGE_VERSION));
        ectlno_end(ex);
        ectlfr_end(fr);
	return TCL_OK;

L_3:    ectlfr_ontrap(fr, L_2);
        for (size_t i = 0; i < sizeof cmd/sizeof cmd[0]; i++)
                if (cmd[i].token)
                        Tcl_DeleteCommandFromToken(interp, cmd[i].token);
L_2:	ectlfr_ontrap(fr, L_1);
	Tcl_Release(interp);
L_1:    ectlfr_ontrap(fr, L_0);
	snmp_shutdown("snmpapp");
L_0:    ectlno_tcl_setresult(interp);
        ectlno_clearmessage();
        ectlno_end(ex);
        ectlfr_end(fr);
        return TCL_ERROR;

}

